@@ -1,3 +1,4 @@
+{-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE OverloadedStrings #-}
@@ -108,7 +109,7 @@ instance Textual LongName where
 
 {- |
 The setup for parsing the command-line arguments of your program. You build
-a @Config@ with 'simple' or 'complex', and pass it to
+a @Config@ with 'simpleConfig' or 'complexConfig', and pass it to
 'Core.Program.Context.configure'.
 -}
 data Config
@@ -127,7 +128,7 @@ data Config
 {- |
 A completely empty configuration, without the default debugging and logging
 options. Your program won't process any command-line options or arguments,
-which would be weird in most cases. Prefer 'simple'.
+which would be weird in most cases. Prefer 'simpleConfig'.
 
 @since 0.2.9
 -}
@@ -171,14 +172,14 @@ with @--help@, would result in:
 \$ __./snippet --help__
 Usage:
 
-    snippet [OPTIONS] filename
+    snippet [OPTIONS] <filename>
 
 Available options:
 
   -h, --host     Specify an alternate host to connect to when performing the
                  frobnication. The default is \"localhost\".
   -p, --port     Specify an alternate port to connect to when frobnicating.
-      --dry-run=TIME
+      --dry-run=<TIME>
                  Perform a trial run at the specified time but don't
                  actually do anything.
   -q, --quiet    Supress normal output.
@@ -188,7 +189,7 @@ Available options:
 
 Required arguments:
 
-  filename       The file you want to frobnicate.
+  <filename>     The file you want to frobnicate.
 \$ __|__
 @
 
@@ -213,7 +214,7 @@ program = ...
 
 main :: 'IO' ()
 main = do
-    context <- 'Core.Program.Execute.configure' ('Core.Program.Execute.fromPackage' version) 'mempty' ('complex'
+    context <- 'Core.Program.Execute.configure' ('Core.Program.Execute.fromPackage' version) 'mempty' ('complexConfig'
         [ 'Global'
             [ 'Option' "station-name" 'Nothing' ('Value' \"NAME\") ['quote'|
                 Specify an alternate radio station to connect to when performing
@@ -323,6 +324,18 @@ By convention a /description/ is one or more complete sentences each of
 which ends with a full stop. For options that take values, use /upper case/
 when specifying the label to be used in help output.
 
+'Remaining' is special; it indicates that you are expecting a variable number
+of additional, non-mandatory arguments. This is used for programs which take a
+list of files to process, for example. It'll show up in the help with the
+description you supply alongside.
+
+@
+        [ ...
+        , 'Remaining' "The files you wish to delete permanently."
+        , ...
+        ]
+@
+
 'Variable' declares an /environment variable/ that, if present, will be
 read by the program and stored in its runtime context. By convention these
 are /upper case/. If the identifier is two or more words they are joined
@@ -338,18 +351,20 @@ with an underscore:
 data Options
     = Option LongName (Maybe ShortName) ParameterValue Description
     | Argument LongName Description
+    | Remaining Description
     | Variable LongName Description
+    deriving (Show)
 
 appendOption :: Options -> Config -> Config
 appendOption option config =
     case config of
         Blank -> Blank
-        Simple options -> Simple (List.reverse (option : options))
+        Simple options -> Simple (options ++ [option])
         Complex commands -> Complex (List.foldl' f [] commands)
   where
     f :: [Commands] -> Commands -> [Commands]
     f acc command = case command of
-        Global options -> Global (List.reverse (option : options)) : acc
+        Global options -> Global (options ++ [option]) : acc
         c@(Command _ _ _) -> c : acc
 
 {- |
@@ -406,6 +421,7 @@ would be parsed as:
 data Parameters = Parameters
     { commandNameFrom :: Maybe LongName
     , parameterValuesFrom :: Map LongName ParameterValue
+    , remainingArgumentsFrom :: [String]
     , environmentValuesFrom :: Map LongName ParameterValue
     }
     deriving (Show, Eq)
@@ -415,6 +431,7 @@ emptyParameters =
     Parameters
         { commandNameFrom = Nothing
         , parameterValuesFrom = emptyMap
+        , remainingArgumentsFrom = []
         , environmentValuesFrom = emptyMap
         }
 
@@ -520,37 +537,40 @@ programName :: String
 programName = unsafePerformIO getProgName
 
 {- |
-Given a program configuration schema and the command-line arguments,
-process them into key/value pairs in a Parameters object.
+Given a program configuration schema and the command-line arguments, process
+them into key/value pairs in a Parameters object.
 
-This results in 'InvalidCommandLine' on the left side if one of the passed
-in options is unrecognized or if there is some other problem handling
-options or arguments (because at that point, we want to rabbit right back
-to the top and bail out; there's no recovering).
+This results in 'InvalidCommandLine' on the left side if one of the passed in
+options is unrecognized or if there is some other problem handling options or
+arguments (because at that point, we want to rabbit right back to the top and
+bail out; there's no recovering).
 
 This isn't something you'll ever need to call directly; it's exposed for
 testing convenience. This function is invoked when you call
 'Core.Program.Context.configure' or 'Core.Program.Execute.execute' (which
-calls 'configure' with a default @Config@ when initializing).
+calls 'Core.Program.Context.configure' with a default 'Config' when
+initializing).
 -}
 parseCommandLine :: Config -> [String] -> Either InvalidCommandLine Parameters
 parseCommandLine config argv = case config of
-    Blank -> return (Parameters Nothing emptyMap emptyMap)
+    Blank -> return (Parameters Nothing emptyMap [] emptyMap)
     Simple options -> do
-        params <- extractor Nothing options argv
-        return (Parameters Nothing params emptyMap)
+        (params, remainder) <- extractor Nothing options argv
+        checkRemainder options remainder
+        return (Parameters Nothing params remainder emptyMap)
     Complex commands ->
         let globalOptions = extractGlobalOptions commands
             modes = extractValidModes commands
          in do
                 (possibles, argv') <- splitCommandLine1 argv
-                params1 <- extractor Nothing globalOptions possibles
-                (first, remainingArgs) <- splitCommandLine2 argv'
+                (params1, _) <- extractor Nothing globalOptions possibles
+                (first, moreArgs) <- splitCommandLine2 argv'
                 (mode, localOptions) <- parseIndicatedCommand modes first
-                params2 <- extractor (Just mode) localOptions remainingArgs
-                return (Parameters (Just mode) ((<>) params1 params2) emptyMap)
+                (params2, remainder) <- extractor (Just mode) localOptions moreArgs
+                checkRemainder localOptions remainder
+                return (Parameters (Just mode) ((<>) params1 params2) remainder emptyMap)
   where
-    extractor :: Maybe LongName -> [Options] -> [String] -> Either InvalidCommandLine (Map LongName ParameterValue)
+    extractor :: Maybe LongName -> [Options] -> [String] -> Either InvalidCommandLine ((Map LongName ParameterValue), [String])
     extractor mode options args =
         let (possibles, arguments) = List.partition isOption args
             valids = extractValidNames options
@@ -558,8 +578,28 @@ parseCommandLine config argv = case config of
             needed = extractRequiredArguments options
          in do
                 list1 <- parsePossibleOptions mode valids shorts possibles
-                list2 <- parseRequiredArguments needed arguments
-                return ((<>) (intoMap list1) (intoMap list2))
+                (list2, arguments') <- parseRequiredArguments needed arguments
+                pure (((<>) (intoMap list1) (intoMap list2)), arguments')
+
+    checkRemainder :: [Options] -> [String] -> Either InvalidCommandLine ()
+    checkRemainder options remainder =
+        if List.null remainder
+            then Right ()
+            else
+                if hasRemaining options
+                    then Right ()
+                    else Left (UnexpectedArguments remainder)
+
+-- is one of the options Remaining?
+hasRemaining :: [Options] -> Bool
+hasRemaining options =
+    List.foldl'
+        ( \acc option -> case option of
+            Remaining _ -> True
+            _ -> acc
+        )
+        False
+        options
 
 isOption :: String -> Bool
 isOption arg = case arg of
@@ -603,21 +643,20 @@ parsePossibleOptions mode valids shorts args = mapM f args
 parseRequiredArguments ::
     [LongName] ->
     [String] ->
-    Either InvalidCommandLine [(LongName, ParameterValue)]
+    Either InvalidCommandLine ([(LongName, ParameterValue)], [String])
 parseRequiredArguments needed argv = iter needed argv
   where
-    iter :: [LongName] -> [String] -> Either InvalidCommandLine [(LongName, ParameterValue)]
-
-    iter [] [] = Right []
+    iter :: [LongName] -> [String] -> Either InvalidCommandLine ([(LongName, ParameterValue)], [String])
+    iter [] [] = Right ([], [])
     -- more arguments supplied than expected
-    iter [] args = Left (UnexpectedArguments args)
+    iter [] args = Right ([], args)
     -- more arguments required, not satisfied
     iter (name : _) [] = Left (MissingArgument name)
     iter (name : names) (arg : args) =
         let deeper = iter names args
          in case deeper of
                 Left e -> Left e
-                Right list -> Right ((name, Value arg) : list)
+                Right (list, remainder) -> Right (((name, Value arg) : list), remainder)
 
 parseIndicatedCommand ::
     Map LongName [Options] ->
@@ -653,11 +692,11 @@ extractShortNames options =
 
 extractRequiredArguments :: [Options] -> [LongName]
 extractRequiredArguments arguments =
-    foldr h [] arguments
+    List.foldl' h [] arguments
   where
-    h :: Options -> [LongName] -> [LongName]
-    h (Argument longname _) needed = longname : needed
-    h _ needed = needed
+    h :: [LongName] -> Options -> [LongName]
+    h needed (Argument longname _) = longname : needed
+    h needed _ = needed
 
 extractGlobalOptions :: [Commands] -> [Options]
 extractGlobalOptions commands =
@@ -669,13 +708,13 @@ extractGlobalOptions commands =
 
 extractValidModes :: [Commands] -> Map LongName [Options]
 extractValidModes commands =
-    foldr k emptyMap commands
+    List.foldl' k emptyMap commands
   where
-    k :: Commands -> Map LongName [Options] -> Map LongName [Options]
-    k (Command longname _ options) modes = insertKeyValue longname options modes
-    k _ modes = modes
+    k :: Map LongName [Options] -> Commands -> Map LongName [Options]
+    k modes (Command longname _ options) = insertKeyValue longname options modes
+    k modes _ = modes
 
-{- |
+{-
 Break the command-line apart in two steps. The first peels off the global
 options, the second below looks to see if there is a command (of fails) and
 if so, whether it has any parameters.
@@ -734,14 +773,14 @@ extractVariableNames options =
 -- The code from here on is formatting code. It's fairly repetative
 -- and crafted to achieve a specific aesthetic output. Rather messy.
 -- I'm sure it could be done "better" but no matter; this is on the
--- path to an exit and return to user's command line.
+-- path to an exit and return to user's shell prompt.
 --
 
 buildUsage :: Config -> Maybe LongName -> Doc ann
 buildUsage config mode = case config of
     Blank -> emptyDoc
     Simple options ->
-        let (o, a) = partitionParameters options
+        let (o, a, v) = partitionParameters options
          in "Usage:" <> hardline <> hardline
                 <> indent
                     4
@@ -751,6 +790,7 @@ buildUsage config mode = case config of
                             [ pretty programName
                             , optionsSummary o
                             , argumentsSummary a
+                            , remainingSummary a
                             ]
                         )
                     )
@@ -759,11 +799,13 @@ buildUsage config mode = case config of
                 <> formatParameters o
                 <> argumentsHeading a
                 <> formatParameters a
+                <> variablesHeading v
+                <> formatParameters v
     Complex commands ->
         let globalOptions = extractGlobalOptions commands
             modes = extractValidModes commands
 
-            (oG, _) = partitionParameters globalOptions
+            (oG, _, vG) = partitionParameters globalOptions
          in "Usage:" <> hardline <> hardline <> case mode of
                 Nothing ->
                     indent
@@ -782,8 +824,10 @@ buildUsage config mode = case config of
                         <> formatParameters oG
                         <> commandHeading modes
                         <> formatCommands commands
+                        <> variablesHeading vG
+                        <> formatParameters vG
                 Just longname ->
-                    let (oL, aL) = case lookupKeyValue longname modes of
+                    let (oL, aL, vL) = case lookupKeyValue longname modes of
                             Just localOptions -> partitionParameters localOptions
                             Nothing -> error "Illegal State"
                      in indent
@@ -796,6 +840,7 @@ buildUsage config mode = case config of
                                     , commandSummary modes
                                     , localSummary oL
                                     , argumentsSummary aL
+                                    , remainingSummary aL
                                     ]
                                 )
                             )
@@ -804,9 +849,11 @@ buildUsage config mode = case config of
                             <> formatParameters oL
                             <> argumentsHeading aL
                             <> formatParameters aL
+                            <> variablesHeading vL
+                            <> formatParameters vL
   where
-    partitionParameters :: [Options] -> ([Options], [Options])
-    partitionParameters options = foldr f ([], []) options
+    partitionParameters :: [Options] -> ([Options], [Options], [Options])
+    partitionParameters options = List.foldl' f ([], [], []) options
 
     optionsSummary :: [Options] -> Doc ann
     optionsSummary os = if length os > 0 then softline <> "[OPTIONS]" else emptyDoc
@@ -831,22 +878,28 @@ buildUsage config mode = case config of
         Nothing -> "COMMAND..."
 
     argumentsSummary :: [Options] -> Doc ann
-    argumentsSummary as = " " <> fillSep (fmap pretty (extractRequiredArguments as))
+    argumentsSummary as = " " <> fillSep (fmap (\x -> "<" <> pretty x <> ">") (extractRequiredArguments as))
 
     argumentsHeading as = if length as > 0 then hardline <> "Required arguments:" <> hardline else emptyDoc
 
+    variablesHeading vs = if length vs > 0 then hardline <> "Known environment variables:" <> hardline else emptyDoc
+
+    remainingSummary :: [Options] -> Doc ann
+    remainingSummary as = if hasRemaining as then " ..." else emptyDoc
+
     -- there is a corner case of complex config with no commands
     commandSummary modes = if length modes > 0 then softline <> commandName else emptyDoc
     commandHeading modes = if length modes > 0 then hardline <> "Available commands:" <> hardline else emptyDoc
 
-    f :: Options -> ([Options], [Options]) -> ([Options], [Options])
-    f o@(Option _ _ _ _) (opts, args) = (o : opts, args)
-    f a@(Argument _ _) (opts, args) = (opts, a : args)
-    f (Variable _ _) (opts, args) = (opts, args)
+    f :: ([Options], [Options], [Options]) -> Options -> ([Options], [Options], [Options])
+    f (opts, args, vars) o@(Option _ _ _ _) = (o : opts, args, vars)
+    f (opts, args, vars) a@(Argument _ _) = (opts, a : args, vars)
+    f (opts, args, vars) a@(Remaining _) = (opts, a : args, vars)
+    f (opts, args, vars) v@(Variable _ _) = (opts, args, v : vars)
 
     formatParameters :: [Options] -> Doc ann
     formatParameters [] = emptyDoc
-    formatParameters options = hardline <> foldr g emptyDoc options
+    formatParameters options = hardline <> List.foldl' g emptyDoc options
 
     --
     -- 16 characters width for short option, long option, and two spaces. If the
@@ -857,8 +910,8 @@ buildUsage config mode = case config of
     -- pretty good and better than waiting until column 8.
     --
 
-    g :: Options -> Doc ann -> Doc ann
-    g (Option longname shortname valued description) acc =
+    g :: Doc ann -> Options -> Doc ann
+    g acc (Option longname shortname valued description) =
         let s = case shortname of
                 Just shortchar -> "  -" <> pretty shortchar <> ", --"
                 Nothing -> "      --"
@@ -868,25 +921,28 @@ buildUsage config mode = case config of
                 Empty ->
                     fillBreak 16 (s <> l <> " ") <+> align (reflow d) <> hardline <> acc
                 Value label ->
-                    fillBreak 16 (s <> l <> "=" <> pretty label <> " ") <+> align (reflow d) <> hardline <> acc
-    g (Argument longname description) acc =
+                    fillBreak 16 (s <> l <> "=<" <> pretty label <> "> ") <+> align (reflow d) <> hardline <> acc
+    g acc (Argument longname description) =
         let l = pretty longname
             d = fromRope description
-         in fillBreak 16 ("  " <> l <> " ") <+> align (reflow d) <> hardline <> acc
-    g (Variable longname description) acc =
+         in fillBreak 16 ("  <" <> l <> "> ") <+> align (reflow d) <> hardline <> acc
+    g acc (Remaining description) =
+        let d = fromRope description
+         in fillBreak 16 ("  " <> "... ") <+> align (reflow d) <> hardline <> acc
+    g acc (Variable longname description) =
         let l = pretty longname
             d = fromRope description
          in fillBreak 16 ("  " <> l <> " ") <+> align (reflow d) <> hardline <> acc
 
     formatCommands :: [Commands] -> Doc ann
-    formatCommands commands = hardline <> foldr h emptyDoc commands
+    formatCommands commands = hardline <> List.foldl' h emptyDoc commands
 
-    h :: Commands -> Doc ann -> Doc ann
-    h (Command longname description _) acc =
+    h :: Doc ann -> Commands -> Doc ann
+    h acc (Command longname description _) =
         let l = pretty longname
             d = fromRope description
-         in fillBreak 16 ("  " <> l <> " ") <+> align (reflow d) <> hardline <> acc
-    h _ acc = acc
+         in acc <> fillBreak 16 ("  " <> l <> " ") <+> align (reflow d) <> hardline
+    h acc _ = acc
 
 buildVersion :: Version -> Doc ann
 buildVersion version =
