@@ -2,6 +2,7 @@
 {-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE DuplicateRecordFields #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE ImportQualifiedPost #-}
 {-# LANGUAGE InstanceSigs #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE OverloadedStrings #-}
@@ -25,7 +26,6 @@ module Core.Program.Context (
     handleTelemetryChoice,
     Exporter (..),
     Forwarder (..),
-    emptyForwarder,
     None (..),
     isNone,
     configure,
@@ -37,28 +37,29 @@ module Core.Program.Context (
     subProgram,
 ) where
 
-import Chrono.TimeStamp (TimeStamp, getCurrentTimeNanoseconds)
 import Control.Concurrent.MVar (MVar, newEmptyMVar, newMVar, putMVar, readMVar)
 import Control.Concurrent.STM.TQueue (TQueue, newTQueueIO)
-import qualified Control.Exception.Safe as Safe (throw)
+import Control.Exception.Safe qualified as Safe (throw)
 import Control.Monad.Catch (MonadCatch, MonadMask, MonadThrow (throwM))
 import Control.Monad.Reader.Class (MonadReader (..))
 import Control.Monad.Trans.Reader (ReaderT (..))
+import Core.Data.Clock
 import Core.Data.Structures
 import Core.Encoding.Json
 import Core.Program.Arguments
 import Core.Program.Metadata
-import Core.System.Base hiding (catch, throw)
+import Core.System.Base
 import Core.Text.Rope
 import Data.Foldable (foldrM)
 import Data.Int (Int64)
 import Data.String (IsString)
 import Prettyprinter (LayoutOptions (..), PageWidth (..), layoutPretty)
 import Prettyprinter.Render.Text (renderIO)
-import qualified System.Console.Terminal.Size as Terminal (Window (..), size)
+import System.Console.Terminal.Size qualified as Terminal (Window (..), size)
 import System.Environment (getArgs, getProgName, lookupEnv)
 import System.Exit (ExitCode (..), exitWith)
-import qualified System.Posix.Process as Posix (exitImmediately)
+import System.IO (hIsTerminalDevice)
+import System.Posix.Process qualified as Posix (exitImmediately)
 import Prelude hiding (log)
 
 {- |
@@ -74,7 +75,7 @@ data Datum = Datum
     { spanIdentifierFrom :: Maybe Span
     , spanNameFrom :: Rope
     , serviceNameFrom :: Maybe Rope
-    , spanTimeFrom :: TimeStamp
+    , spanTimeFrom :: Time
     , traceIdentifierFrom :: Maybe Trace
     , parentIdentifierFrom :: Maybe Span
     , durationFrom :: Maybe Int64
@@ -88,7 +89,7 @@ emptyDatum =
         { spanIdentifierFrom = Nothing
         , spanNameFrom = emptyRope
         , serviceNameFrom = Nothing
-        , spanTimeFrom = 0
+        , spanTimeFrom = epochTime
         , traceIdentifierFrom = Nothing
         , parentIdentifierFrom = Nothing
         , durationFrom = Nothing
@@ -102,7 +103,7 @@ continuing an inherited trace and passed the identifier of the parent span you
 can specify it using this constructor.
 -}
 newtype Span = Span Rope
-    deriving (Show, IsString)
+    deriving (Show, Eq, IsString)
 
 unSpan :: Span -> Rope
 unSpan (Span text) = text
@@ -116,7 +117,7 @@ this program or request handler, and you can specify it by using
 'Core.Telemetry.Observability.usingTrace'.
 -}
 newtype Trace = Trace Rope
-    deriving (Show, IsString)
+    deriving (Show, Eq, IsString)
 
 unTrace :: Trace -> Rope
 unTrace (Trace text) = text
@@ -134,12 +135,6 @@ data Forwarder = Forwarder
     { telemetryHandlerFrom :: [Datum] -> IO ()
     }
 
-emptyForwarder :: Forwarder
-emptyForwarder =
-    Forwarder
-        { telemetryHandlerFrom = \_ -> pure ()
-        }
-
 {- |
 Internal context for a running program. You access this via actions in the
 'Program' monad. The principal item here is the user-supplied top-level
@@ -167,24 +162,19 @@ application data of type @τ@ which can be retrieved with
 -- that field name as a local variable name.
 --
 data Context τ = Context
-    { -- runtime properties
-      programNameFrom :: MVar Rope
+    { programNameFrom :: MVar Rope
     , terminalWidthFrom :: Int
+    , terminalColouredFrom :: Bool
     , versionFrom :: Version
-    , -- only used during initial setup
-      initialConfigFrom :: Config
+    , initialConfigFrom :: Config -- only used during initial setup
     , initialExportersFrom :: [Exporter]
-    , -- derived at startup
-      commandLineFrom :: Parameters
-    , -- operational state
-      exitSemaphoreFrom :: MVar ExitCode
-    , startTimeFrom :: MVar TimeStamp
+    , commandLineFrom :: Parameters -- derived at startup
+    , exitSemaphoreFrom :: MVar ExitCode
+    , startTimeFrom :: MVar Time
     , verbosityLevelFrom :: MVar Verbosity
-    , -- communication channels
-      outputChannelFrom :: TQueue (Maybe Rope)
-    , telemetryChannelFrom :: TQueue (Maybe Datum)
-    , -- machinery for telemetry
-      telemetryForwarderFrom :: Forwarder
+    , outputChannelFrom :: TQueue (Maybe Rope) -- communication channels
+    , telemetryChannelFrom :: TQueue (Maybe Datum) -- machinery for telemetry
+    , telemetryForwarderFrom :: Maybe Forwarder
     , currentDatumFrom :: MVar Datum
     , applicationDataFrom :: MVar τ
     }
@@ -237,6 +227,8 @@ data Verbosity
     | -- | @since 0.2.12
       Verbose
     | Debug
+    | -- | @since 0.4.6
+      Internal
     deriving (Show)
 
 {- |
@@ -294,7 +286,14 @@ to another package's entry point, just use 'liftIO'. It's re-exported by
 unlifting back to Program; see "Core.Program.Unlift".
 -}
 newtype Program τ α = Program (ReaderT (Context τ) IO α)
-    deriving (Functor, Applicative, Monad, MonadIO, MonadReader (Context τ))
+    deriving
+        ( Functor
+        , Applicative
+        , Monad
+        , MonadIO
+        , MonadReader (Context τ)
+        , MonadFail
+        )
 
 unProgram :: Program τ α -> ReaderT (Context τ) IO α
 unProgram (Program r) = r
@@ -352,6 +351,7 @@ configure version t config = do
     q <- newEmptyMVar
     i <- newMVar start
     columns <- getConsoleWidth
+    coloured <- getConsoleColoured
     level <- newEmptyMVar
     out <- newTQueueIO
     tel <- newTQueueIO
@@ -363,6 +363,7 @@ configure version t config = do
         $! Context
             { programNameFrom = n
             , terminalWidthFrom = columns
+            , terminalColouredFrom = coloured
             , versionFrom = version
             , initialConfigFrom = config
             , initialExportersFrom = []
@@ -372,7 +373,7 @@ configure version t config = do
             , verbosityLevelFrom = level -- will be filled in handleVerbosityLevel
             , outputChannelFrom = out
             , telemetryChannelFrom = tel
-            , telemetryForwarderFrom = emptyForwarder
+            , telemetryForwarderFrom = Nothing
             , currentDatumFrom = v
             , applicationDataFrom = u
             }
@@ -391,7 +392,10 @@ getConsoleWidth = do
             Nothing -> 80
     return columns
 
---
+getConsoleColoured :: IO Bool
+getConsoleColoured = do
+    terminal <- hIsTerminalDevice stdout
+    pure terminal
 
 {- |
 Process the command line options and arguments. If an invalid option is
@@ -458,7 +462,7 @@ lookupEnvironmentVariables config params = do
         result <- lookupEnv var
         return $ case result of
             Just value -> insertKeyValue name (Value value) acc
-            Nothing -> acc
+            Nothing -> insertKeyValue name Empty acc
 
 handleVerbosityLevel :: Context τ -> IO (MVar Verbosity)
 handleVerbosityLevel context = do
@@ -481,6 +485,7 @@ queryVerbosityLevel params =
      in case debug of
             Just value -> case value of
                 Empty -> Right Debug
+                Value "internal" -> Right Internal
                 Value _ -> Left (ExitFailure 2)
             Nothing -> case verbose of
                 Just value -> case value of
@@ -514,7 +519,7 @@ handleTelemetryChoice context = do
                 -- and return it
                 pure
                     context
-                        { telemetryForwarderFrom = forwarder
+                        { telemetryForwarderFrom = Just forwarder
                         }
   where
     lookupExporter :: Rope -> [Exporter] -> Maybe Exporter
