@@ -1,4 +1,5 @@
 {-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE ImportQualifiedPost #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RankNTypes #-}
@@ -61,10 +62,12 @@ module Core.Program.Execute (
 
     -- * Accessing program context
     getCommandLine,
-    lookupOptionFlag,
-    lookupOptionValue,
-    lookupArgument,
-    lookupEnvironmentValue,
+    queryCommandName,
+    queryOptionFlag,
+    queryOptionValue,
+    queryArgument,
+    queryRemaining,
+    queryEnvironmentValue,
     getProgramName,
     setProgramName,
     getVerbosityLevel,
@@ -77,43 +80,43 @@ module Core.Program.Execute (
     outputEntire,
     inputEntire,
     execProcess,
-
-    -- * Concurrency
-    Thread,
-    forkThread,
     sleepThread,
     resetTimer,
-    waitThread,
-    waitThread_,
     trap_,
 
+    -- * Exception handling
+    catch,
+    throw,
+    try,
+
     -- * Internals
     Context,
     None (..),
     isNone,
     unProgram,
-    unThread,
     invalid,
+    Boom (..),
     loopForever,
+    lookupOptionFlag,
+    lookupOptionValue,
+    lookupArgument,
+    lookupEnvironmentValue,
 ) where
 
-import Chrono.TimeStamp (getCurrentTimeNanoseconds)
 import Control.Concurrent (threadDelay)
 import Control.Concurrent.Async (
-    Async,
     ExceptionInLinkedThread (..),
  )
-import qualified Control.Concurrent.Async as Async (
+import Control.Concurrent.Async qualified as Async (
     async,
     cancel,
-    link,
     race,
     race_,
     wait,
  )
 import Control.Concurrent.MVar (
+    MVar,
     modifyMVar_,
-    newMVar,
     putMVar,
     readMVar,
  )
@@ -127,28 +130,46 @@ import Control.Concurrent.STM.TQueue (
     unGetTQueue,
     writeTQueue,
  )
-import qualified Control.Exception as Base (throwIO)
-import qualified Control.Exception.Safe as Safe (catch, catchesAsync, throw)
+import Control.Exception qualified as Base (throwIO)
+import Control.Exception.Safe qualified as Safe (
+    catch,
+    catchesAsync,
+    throw,
+ )
 import Control.Monad (
     void,
     when,
  )
 import Control.Monad.Catch (Handler (..))
 import Control.Monad.Reader.Class (MonadReader (ask))
+import Core.Data.Clock
 import Core.Data.Structures
 import Core.Program.Arguments
 import Core.Program.Context
+import Core.Program.Exceptions
 import Core.Program.Logging
 import Core.Program.Signal
-import Core.System.Base
+import Core.System.Base (
+    Exception,
+    Handle,
+    SomeException,
+    displayException,
+    hFlush,
+    liftIO,
+    stdout,
+ )
 import Core.Text.Bytes
 import Core.Text.Rope
-import qualified Data.ByteString as B (hPut)
-import qualified Data.ByteString.Char8 as C (singleton)
+import Data.ByteString qualified as B (hPut)
+import Data.ByteString.Char8 qualified as C (singleton)
+import Data.List qualified as List (intersperse)
 import GHC.Conc (getNumProcessors, numCapabilities, setNumCapabilities)
 import GHC.IO.Encoding (setLocaleEncoding, utf8)
+import System.Directory (
+    findExecutable,
+ )
 import System.Exit (ExitCode (..))
-import qualified System.Posix.Process as Posix (exitImmediately)
+import System.Posix.Process qualified as Posix (exitImmediately)
 import System.Process.Typed (closed, proc, readProcess, setStdin)
 import Prelude hiding (log)
 
@@ -198,9 +219,9 @@ Trap any exceptions coming out of the given Program action, and discard them.
 The one and only time you want this is inside an endless loop:
 
 @
-    forever $ do
-        trap_
-            ( bracket
+    'Conrol.Monad.forever' $ do
+        'trap_'
+            ( 'bracket'
                 obtainResource
                 releaseResource
                 useResource
@@ -214,6 +235,8 @@ released resources with 'bracket' or 'finally' as shown above.
 An info level message will be sent to the log channel indicating that an
 uncaught exception was trapped along with a debug level message showing the
 exception text, if any.
+
+@since 0.2.11
 -}
 trap_ :: Program τ α -> Program τ ()
 trap_ action =
@@ -275,7 +298,7 @@ executeActual context0 program = do
     -- set up debug logger
     l <-
         Async.async $ do
-            processTelemetryMessages forwarder out tel
+            processTelemetryMessages forwarder level out tel
 
     -- run actual program, ensuring to grab any otherwise uncaught exceptions.
     code <-
@@ -306,9 +329,12 @@ executeActual context0 program = do
         ( do
             atomically $ do
                 writeTQueue tel Nothing
-                writeTQueue out Nothing
 
             Async.wait l
+
+            atomically $ do
+                writeTQueue out Nothing
+
             Async.wait o
         )
         ( do
@@ -351,42 +377,54 @@ processStandardOutput out =
 -- the technique used   by **io-streams** to just pass along a stream of Maybes,
 -- with Nothing signalling end-of-stream is exactly good enough for our needs.
 --
-processTelemetryMessages :: Forwarder -> TQueue (Maybe Rope) -> TQueue (Maybe Datum) -> IO ()
-processTelemetryMessages processor out tel = do
+processTelemetryMessages :: Maybe Forwarder -> MVar Verbosity -> TQueue (Maybe Rope) -> TQueue (Maybe Datum) -> IO ()
+processTelemetryMessages Nothing _ _ tel = do
+    ignoreForever tel
+  where
+    ignoreForever queue = do
+        possibleItem <- atomically $ do
+            readTQueue queue -- blocks
+        case possibleItem of
+            -- time to shutdown
+            Nothing -> pure ()
+            -- otherwise igonore
+            Just _ -> do
+                ignoreForever queue
+processTelemetryMessages (Just processor) v out tel = do
     Safe.catch
-        (loopForever action out tel)
+        (loopForever action v out tel)
         (collapseHandler "telemetry processing collapsed")
   where
     action = telemetryHandlerFrom processor
 
-loopForever :: ([a] -> IO ()) -> TQueue (Maybe Rope) -> TQueue (Maybe a) -> IO ()
-loopForever action out queue = do
-    start <- getCurrentTimeNanoseconds
+loopForever :: ([a] -> IO ()) -> MVar Verbosity -> TQueue (Maybe Rope) -> TQueue (Maybe a) -> IO ()
+loopForever action v out queue = do
     -- block waiting for an item
     possibleItems <- atomically $ do
-        cycleOverQueue []
+        cycleOverQueue 0 []
 
     case possibleItems of
         -- we're done!
         Nothing -> pure ()
         -- handle it and loop
         Just items -> do
-            catch
-                (action (reverse items))
-                ( \(e :: SomeException) -> do
-                    now <- getCurrentTimeNanoseconds
-                    let result =
-                            formatLogMessage
-                                start
-                                now
-                                SeverityWarn
-                                ("sending telemetry failed (Exception: " <> intoRope (show e) <> "); Restarting exporter.")
-                    atomically $ do
-                        writeTQueue out (Just result)
+            start <- getCurrentTimeNanoseconds
+            Safe.catch
+                ( do
+                    action (reverse items)
+                    reportStatus start (length items)
                 )
-            loopForever action out queue
+                ( \(e :: SomeException) -> do
+                    reportProblem start e
+                )
+            loopForever action v out queue
   where
-    cycleOverQueue items =
+    cycleOverQueue !count items =
+        if count >= (1024 :: Int)
+            then pure (Just items)
+            else cycleOverQueue' count items
+
+    cycleOverQueue' !count items =
         case items of
             [] -> do
                 possibleItem <- readTQueue queue -- blocks
@@ -395,7 +433,7 @@ loopForever action out queue = do
                     Nothing -> pure Nothing
                     -- otherwise start accumulating
                     Just item -> do
-                        cycleOverQueue (item : [])
+                        cycleOverQueue 1 (item : [])
             _ -> do
                 pending <- tryReadTQueue queue -- doesn't block
                 case pending of
@@ -413,7 +451,38 @@ loopForever action out queue = do
                                 pure (Just items)
                             -- continue accumulating!
                             Just item -> do
-                                cycleOverQueue (item : items)
+                                cycleOverQueue (count + 1) (item : items)
+
+    reportStatus start num = do
+        level <- readMVar v
+        when (isInternal level) $ do
+            now <- getCurrentTimeNanoseconds
+            let desc = case num of
+                    1 -> "1 event"
+                    _ -> intoRope (show num) <> " events"
+                message =
+                    formatLogMessage
+                        start
+                        now
+                        True
+                        SeverityInternal
+                        ("Sent " <> desc)
+            atomically $ do
+                writeTQueue out (Just message)
+
+    reportProblem start e = do
+        level <- readMVar v
+        when (isEvent level) $ do
+            now <- getCurrentTimeNanoseconds
+            let message =
+                    formatLogMessage
+                        start
+                        now
+                        True
+                        SeverityWarn
+                        ("Sending telemetry failed (Exception: " <> intoRope (show e) <> "); Restarting exporter.")
+            atomically $ do
+                writeTQueue out (Just message)
 
 {- |
 Safely exit the program with the supplied exit code. Current output and
@@ -445,7 +514,7 @@ getVerbosityLevel = do
 
 {- |
 Change the verbosity level of the program's logging output. This changes
-whether 'event' and the 'debug' family of functions emit to the logging
+whether 'info' and the 'debug' family of functions emit to the logging
 stream; they do /not/ affect 'write'ing to the terminal on the standard
 output stream.
 -}
@@ -482,7 +551,7 @@ getProgramName = do
 Retreive the current terminal's width, in characters.
 
 If you are outputting an object with a 'Core.Text.Untilities.Render'
-instance then you may not need this; you can instead use 'wrteR' which is
+instance then you may not need this; you can instead use 'writeR' which is
 aware of the width of your terminal and will reflow (in as much as the
 underlying type's @Render@ instance lets it).
 -}
@@ -498,7 +567,7 @@ Get the user supplied application state as originally supplied to
 'setApplicationState'.
 
 @
-    state <- getApplicationState
+    settings <- 'getApplicationState'
 @
 -}
 getApplicationState :: Program τ τ
@@ -512,8 +581,8 @@ getApplicationState = do
 Update the user supplied top-level application state.
 
 @
-    let state' = state { answer = 42 }
-    setApplicationState state'
+    let settings' = settings { answer = 42 }
+    'setApplicationState' settings'
 @
 -}
 setApplicationState :: τ -> Program τ ()
@@ -528,11 +597,11 @@ Write the supplied @Bytes@ to the given @Handle@. Note that in contrast to
 'write' we don't output a trailing newline.
 
 @
-    'output' h b
+    'outputEntire' h b
 @
 
 Do /not/ use this to output to @stdout@ as that would bypass the mechanism
-used by the 'write'*, 'event', and 'debug'* functions to sequence output
+used by the 'write'*, 'info', and 'debug'* functions to sequence output
 correctly. If you wish to write to the terminal use:
 
 @
@@ -551,6 +620,12 @@ Read the (entire) contents of the specified @Handle@.
 inputEntire :: Handle -> Program τ Bytes
 inputEntire handle = liftIO (hInput handle)
 
+data ProcessProblem
+    = CommandNotFound Rope
+    deriving (Show)
+
+instance Exception ProcessProblem
+
 {- |
 Execute an external child process and wait for its output and result. The
 command is specified first and and subsequent arguments as elements of the
@@ -578,54 +653,24 @@ to use something like __io-streams__ instead.
 execProcess :: [Rope] -> Program τ (ExitCode, Rope, Rope)
 execProcess [] = error "No command provided"
 execProcess (cmd : args) =
-    let cmdStr = fromRope cmd
-        argsStr = fromRope <$> args
-        task = proc cmdStr argsStr
-        task' = setStdin closed task
+    let cmd' = fromRope cmd
+        args' = fmap fromRope args
+        task = proc cmd' args'
+        task1 = setStdin closed task
+        command = mconcat (List.intersperse (singletonRope ' ') (cmd : args))
      in do
-            debugS "command" task'
+            debug "command" command
 
-            (exit, out, err) <- liftIO $ do
-                readProcess task'
+            probe <- liftIO $ do
+                findExecutable cmd'
+            case probe of
+                Nothing -> do
+                    Safe.throw (CommandNotFound cmd)
+                Just _ -> do
+                    (exit, out, err) <- liftIO $ do
+                        readProcess task1
 
-            return (exit, intoRope out, intoRope err)
-
-{- |
-A thread for concurrent computation. Haskell uses green threads: small lines
-of work that are scheduled down onto actual execution contexts, set by default
-by this library to be one per core. They are incredibly lightweight, and you
-are encouraged to use them freely. Haskell provides a rich ecosystem of tools
-to do work concurrently and to communicate safely between threads
-
-(this wraps __async__'s 'Async')
--}
-newtype Thread α = Thread (Async α)
-
-unThread :: Thread α -> Async α
-unThread (Thread a) = a
-
-{- |
-Fork a thread. The child thread will run in the same @Context@ as the calling
-@Program@, including sharing the user-defined application state type.
-
-(this wraps __async__'s 'async' which in turn wraps __base__'s
-'Control.Concurrent.forkIO')
--}
-forkThread :: Program τ α -> Program τ (Thread α)
-forkThread program = do
-    context <- ask
-    let i = startTimeFrom context
-
-    liftIO $ do
-        start <- readMVar i
-        i' <- newMVar start
-
-        let context' = context{startTimeFrom = i'}
-
-        a <- Async.async $ do
-            subProgram context' program
-        Async.link a
-        return (Thread a)
+                    pure (exit, intoRope out, intoRope err)
 
 {- |
 Reset the start time (used to calculate durations shown in event- and
@@ -638,13 +683,15 @@ of the total elapsed program time, then fork a new thread for your worker and
 reset the timer there.
 
 @
-    'forkThread' $ do
+    'Core.Program.Threads.forkThread' $ do
         'resetTimer'
         ...
 @
 
 then times output in the log messages will be relative to that call to
 'resetTimer', not the program start.
+
+@since 0.2.7
 -}
 resetTimer :: Program τ ()
 resetTimer = do
@@ -674,40 +721,6 @@ sleepThread seconds =
     let us = floor (toRational (seconds * 1e6))
      in liftIO $ threadDelay us
 
-{- |
-Wait for the completion of a thread, returning the result. This is a blocking
-operation.
-
-(this wraps __async__'s 'wait')
--}
-waitThread :: Thread α -> Program τ α
-waitThread (Thread a) = liftIO $ Async.wait a
-
-{- |
-Wait for the completion of a thread, discarding its result. This is
-particularly useful at the end of a do-block if you're waiting on a worker
-thread to finish but don't need its return value, if any; otherwise you have
-to explicily deal with the unused return value:
-
-@
-    _ <- 'waitThread' t1
-    'return' ()
-@
-
-which is a bit tedious. Instead, you can just use this convenience function:
-
-@
-    'waitThread_' t1
-@
-
-The trailing underscore in the name of this function follows the same
-convetion as found in "Control.Monad", which has 'Control.Monad.mapM_' which
-does the same as 'Control.Monad.mapM' but which likewise discards the return
-value.
--}
-waitThread_ :: Thread α -> Program τ ()
-waitThread_ = void . waitThread
-
 {- |
 Retrieve the values of parameters parsed from options and arguments supplied
 by the user on the command-line.
@@ -734,8 +747,7 @@ the parameter have a value?"
 
 This is available should you need to differentiate between a @Value@ and an
 @Empty@ 'ParameterValue', but for many cases as a convenience you can use the
-'lookupOptionFlag', 'lookupOptionValue', and 'lookupArgument' functions below
-(which are just wrappers around a code block like the example shown here).
+'queryOptionFlag', 'queryOptionValue', and 'queryArgument' functions below.
 -}
 getCommandLine :: Program τ (Parameters)
 getCommandLine = do
@@ -744,11 +756,26 @@ getCommandLine = do
 
 {- |
 Arguments are mandatory, so by the time your program is running a value
-has already been identified. This returns the value for that parameter.
+has already been identified. This retreives the value for that parameter.
+
+@
+program = do
+    file <- 'queryArgument' \"filename\"
+    ...
+@
+
+@since 0.2.7
 -}
+queryArgument :: LongName -> Program τ Rope
+queryArgument name = do
+    context <- ask
+    let params = commandLineFrom context
+    case lookupKeyValue name (parameterValuesFrom params) of
+        Nothing -> error "Attempted lookup of unconfigured argument"
+        Just argument -> case argument of
+            Empty -> error "Invalid State"
+            Value value -> pure (intoRope value)
 
--- this is Maybe because you can inadvertently ask for an unconfigured name
--- this could be fixed with a much stronger Config type, potentially.
 lookupArgument :: LongName -> Parameters -> Maybe String
 lookupArgument name params =
     case lookupKeyValue name (parameterValuesFrom params) of
@@ -756,13 +783,52 @@ lookupArgument name params =
         Just argument -> case argument of
             Empty -> error "Invalid State"
             Value value -> Just value
+{-# DEPRECATED lookupArgument "Use queryArgument instead" #-}
+
+{- |
+In other applications, you want to gather up the remaining arguments on the
+command-line. You need to have specified 'Remaining' in the configuration.
+
+@
+program = do
+    files \<- 'queryRemaining'
+    ...
+@
+
+@since 0.3.5
+-}
+queryRemaining :: Program τ [Rope]
+queryRemaining = do
+    context <- ask
+    let params = commandLineFrom context
+    let remaining = remainingArgumentsFrom params
+    pure (fmap intoRope remaining)
 
 {- |
 Look to see if the user supplied a valued option and if so, what its value
-was.
+was. Use of the @LambdaCase@ extension might make accessing the parameter a
+bit eaiser:
+
+@
+program = do
+    count \<- 'queryOptionValue' \"count\" '>>=' \\case
+        'Nothing' -> 'pure' 0
+        'Just' value -> 'pure' value
+    ...
+@
+
+@since 0.3.5
 -}
+queryOptionValue :: LongName -> Program τ (Maybe Rope)
+queryOptionValue name = do
+    context <- ask
+    let params = commandLineFrom context
+    case lookupKeyValue name (parameterValuesFrom params) of
+        Nothing -> pure Nothing
+        Just argument -> case argument of
+            Empty -> pure (Just emptyRope)
+            Value value -> pure (Just (intoRope value))
 
--- Should this be more severe if it encounters Empty?
 lookupOptionValue :: LongName -> Parameters -> Maybe String
 lookupOptionValue name params =
     case lookupKeyValue name (parameterValuesFrom params) of
@@ -770,24 +836,51 @@ lookupOptionValue name params =
         Just argument -> case argument of
             Empty -> Nothing
             Value value -> Just value
+{-# DEPRECATED lookupOptionValue "Use queryOptionValue instead" #-}
 
 {- |
-Returns @Just True@ if the option is present, and @Nothing@ if it is not.
+Returns @True@ if the option is present, and @False@ if it is not.
+
+@
+program = do
+    overwrite \<- 'queryOptionFlag' \"overwrite\"
+    ...
+@
+
+@since 0.3.5
 -}
+queryOptionFlag :: LongName -> Program τ Bool
+queryOptionFlag name = do
+    context <- ask
+    let params = commandLineFrom context
+    case lookupKeyValue name (parameterValuesFrom params) of
+        Nothing -> pure False
+        Just _ -> pure True
 
--- The type is boolean to support a possible future extension of negated
--- arguments.
 lookupOptionFlag :: LongName -> Parameters -> Maybe Bool
 lookupOptionFlag name params =
     case lookupKeyValue name (parameterValuesFrom params) of
         Nothing -> Nothing
         Just argument -> case argument of
             _ -> Just True -- nom, nom
+{-# DEPRECATED lookupOptionFlag "Use queryOptionFlag instead" #-}
 
 {- |
 Look to see if the user supplied the named environment variable and if so,
 return what its value was.
+
+@since 0.3.5
 -}
+queryEnvironmentValue :: LongName -> Program τ (Maybe Rope)
+queryEnvironmentValue name = do
+    context <- ask
+    let params = commandLineFrom context
+    case lookupKeyValue name (environmentValuesFrom params) of
+        Nothing -> error "Attempted lookup of unconfigured environment variable"
+        Just param -> case param of
+            Empty -> pure Nothing
+            Value str -> pure (Just (intoRope str))
+
 lookupEnvironmentValue :: LongName -> Parameters -> Maybe String
 lookupEnvironmentValue name params =
     case lookupKeyValue name (environmentValuesFrom params) of
@@ -795,6 +888,25 @@ lookupEnvironmentValue name params =
         Just param -> case param of
             Empty -> Nothing
             Value str -> Just str
+{-# DEPRECATED lookupEnvironmentValue "Use queryEnvironment instead" #-}
+
+{- |
+Retreive the sub-command mode selected by the user. This assumes your program
+was set up to take sub-commands via 'complexConfig'.
+
+@
+    mode <- queryCommandName
+@
+
+@since 0.3.5
+-}
+queryCommandName :: Program τ Rope
+queryCommandName = do
+    context <- ask
+    let params = commandLineFrom context
+    case commandNameFrom params of
+        Just (LongName name) -> pure (intoRope name)
+        Nothing -> error "Attempted lookup of command but not a Complex Config"
 
 {- |
 Illegal internal state resulting from what should be unreachable code or
