@@ -28,26 +28,13 @@ module Core.System.Base (
 
     -- * Exception handling
 
-    -- ** from Control.Exception.Safe
-
-    -- | Re-exported from "Control.Exception.Safe" in the __safe-exceptions__ package:
     Exception (..),
     SomeException,
-    throw,
-    impureThrow,
-    bracket,
-    catch,
-    finally,
 ) where
 
-import Control.Exception.Safe (
+import Control.Exception (
     Exception (..),
     SomeException,
-    bracket,
-    catch,
-    finally,
-    impureThrow,
-    throw,
  )
 import Control.Monad.IO.Class (MonadIO, liftIO)
 import System.IO (Handle, IOMode (..), hFlush, stderr, stdin, stdout, withFile)
